% geq_opt.m
% Optimization of the orders of a high-order graphical equalizer

% Created: Jussi Rm and Vesa Vlimki, January 11, 2013
% Last modified: 8 August 2013

% Usage: 
% N_m: Number of the 4th-order block per band  
% bands: 1= Bark (25 bands), 2 = 1/3 octave (29 bands), 3 = octave (10 bands)

clear;
fs = 44100;     % Sampling rate

% Design parameters (Change these to set the design goals)
bands = 1;      % bands: 1= Bark (25 bands), 2 = 1/3 octave (29 bands), 3 = octave (10 bands)
m_max = 20;     % Maximum number of 4th-order sections, max order N_max = 4*m_max = 2*M_max
i_start = 9;    % Start band
G_target = -20; % Target gain
E_target = 2;   % Target error


% Initialize variables
[h,f_M,N_b] = ho_geq(fs,bands,zeros(29,1),zeros(29,1));
% h = filter's impulse response
% f_M = optimised center frequencies (Hz)
% N_b = number of bands

N_m = ones(N_b,1); % Initialize every band to have one 4th-order block

rep = 1;
disp('-------------------')

%----- Automatic minOrder calculation -----%
while(rep ==1 && N_m(i_start) < m_max)
    %------ Step 1) -----%
    g = zeros(N_b,1);                               % Initializes the gain vector
    g(i_start-1:i_start) = G_target*ones(2,1);      % Sets target gain to bands i_start-1 and i_start 
    h = ho_geq(fs,bands,g,N_m,1);                   % Calculates the impulse response of the filter
    % Magnitude spectrum of the filter
    [sp,dfreq] = spect(h,fs);                       % spect.m function
    len = length(sp);                               % Length of the spectrum
    frq = linspace(0,dfreq*(len-1),len);            % Frequency axis
    %----- Step 2) -----%
    % Error limits (center frequencies of both filters)
    llim = find(frq <= f_M(i_start-1), 1 ,'last');
    hlim = find(frq >= f_M(i_start), 1, 'first');
    %----- Step 3) -----%
    % Error
    error = sp(llim:hlim)-G_target;
    E_max = max(abs(error));                        % Peak error
    %----- Step 4) -----%
    if(E_max > E_target)
        N_m(i_start-1) = N_m(i_start-1)+1;          % Add one 4-th order section
        N_m(i_start) = N_m(i_start)+1;              % Add one 4-th order section
        rep = 1;                                    % Repeat loop
    else % if the target is met, save the order
        rep = 0;                                    % Exit loop
        er = E_max;                                 % Save the error value
    end
end

%----- Low frequency filters (i < i_start) -----%
for ind = 1:i_start-1
    %------ Step 1) -----%
    g = zeros(N_b,1);                                       % Initializes gain vector
    g(i_start-ind:i_start-ind+1) = G_target*ones(2,1);      % Sets target gains to the band under evaluation
    prev = abs(G_target);                                   % Previous error-value initialization. This should be
                                                            % big enough that the first error in smaller than this
    while(rep ==1 && N_m(i_start-ind) < m_max)
        h = ho_geq(fs,bands,g,N_m,1);                       % Calculates the impulse response of the filter
        % Magnitude spectrum of the filter
        [sp,dfreq] = spect(h,fs);                           % spect.m function
        %----- Step 2) -----%   
        % Error limits (center frequencies of both filters)
        llim = find(frq <= f_M(i_start-ind), 1 ,'last');
        hlim = find(frq >= f_M(i_start-ind+1), 1, 'first');
        %----- Step 3) -----%
        % Error
        error = sp(llim:hlim)-G_target;
        E_max = max(abs(error));                            % Peak error?
        %----- Step 5) -----%
        % If the error is bigger than the target and smaller
        % than the previous error, increse the order
        if(E_max > E_target && E_max < prev)                
            N_m(i_start-ind) = N_m(i_start-ind)+1;          % Increase the order by adding one 4-th order section
            rep = 1;                                        % Continue the loop
            disp(['Calculating.. Error = ' num2str(E_max,3) ' dB']) % Display current error value
        % If the error is larger than that of the previous order,
        % select the previous order and exit the loop.    
        elseif(E_max > prev) 
            N_m(i_start-ind) = N_m(i_start-ind)-1;          % Remove one 4-th order section
            rep = 0;                                        % Exit the loop
            er = prev;                                      % Save the previous error value to be the final error
        % If the target is met, save the order
        else 
            rep = 0;                                        % Exit the loop
            er = E_max;                                     % Save the error value
        end
        prev = E_max;                                       % Save the calculated error to be the previous error
    end
    errors(ind) = er;                                       % Write final error to a vector
    disp(['FINAL ERROR = ' num2str(er,3) ' dB'])            % Display the final error
    rep = 1;                                                % Set rep = 1 for the next band
end

%----- High frequency filters (i > i_start) -----%
for ind = i_start:N_b-1
    %------ Step 1) -----%
    g = zeros(N_b,1);                                       % Initializes gain vector
    g(ind:ind+1) = G_target*ones(2,1);                      % Sets target gains to the band under evaluation
    prev = abs(G_target);                                   % Previous error-value initialization. This should be
    while(rep ==1 && N_m(ind+1) < m_max)
        h = ho_geq(fs,bands,g,N_m,1);                       % Calculates the impulse response of the filter
        % Spectrum of the filter
        [sp,dfreq] = spect(h,fs);                           % spect.m function
        %----- Step 2) -----%
        % Error limits (center frequencies of both filters)
        llim = find(frq <= f_M(ind), 1 ,'last');
        hlim = find(frq >= f_M(ind+1), 1, 'first');
        %----- Step 3) -----%
        % Error
        error = sp(llim:hlim)-G_target;
        E_max = max(abs(error));
        %----- Step 5) -----%
        % If the error is bigger than the target and smaller
        % than the previous error, increse the order
        if(E_max > E_target && E_max < prev)
            N_m(ind+1) = N_m(ind+1)+1;                      % Increase the order by adding one 4-th order section
            rep = 1;                                        % Continue the loop
            disp(['Calculating.. Error = ' num2str(E_max,3) ' dB']) % Display current error value
        % If the error is larger than that of the previous order,
        % select the previous order and exit the loop.
        elseif(E_max > prev)
            N_m(ind+1) = N_m(ind+1)-1;                      % Remove one 4-th order section
            rep = 0;                                        % Exit the loop
            er = prev;                                      % Save the previous error value to be the final error
        % If the target is met, save the order
        else
            rep = 0;                                        % Exit the loop
            er = E_max;                                     % Save the error value
        end
        prev = E_max;                                       % Save the calculated error to be the previous error
    end
    errors(ind) = er;                                       % Write final error to a vector
    disp(['FINAL ERROR = ' num2str(er,3) ' dB'])            % Display the final error
    rep = 1;                                                % Set rep = 1 for the next band
end

% Calculate the variable M and the actual order of the filters
M = 2.*N_m;
orders = 4.*N_m;

disp('-------------')
disp('-------------')
disp(['Optimized orders: ' num2str(orders')])
disp('-------------')

% Plot the final optimized filter
figure(1)
h = ho_geq(fs,bands,G_target*ones(N_b,1),N_m,1);
hold on;
% Target error limits
semilogx([20 22000], [-E_target+G_target -E_target+G_target],'k') 
semilogx([20 22000], [E_target+G_target E_target+G_target],'k')
hold off;