function [out,f_M,N_b] = ho_geq(fs,bands,g,N_m,plotcode)

% ho_geq.m
%
% High-order Graphic EQ
%
% Ref. M. Holters and U. Zlzer, "Graphic EQ design using
% higher-order recursive filters," in Proc. DAFx-06, Montreal,
% pp. 37-40. Available online at http://www.dafx.ca/
%
% Created: Jussi Rm and Vesa Vlimki, January 9, 2013
% Last modified: 31 July 2013

% Usage: 

% INPUTS:
% fs: Sampling rate
% bands: 1 = Bark (25 bands), 2 = 1/3 octave (29 bands), 3 = octave (10 bands)
% g: Gain vector (dB)
% N_m: Number of the 4th-order blocks per band 
% plotcode: 1 = plot, 0 = don't plot

% OUTPUTS
% out: Impulse response of the filter
% f_M: Optimised center frequencies (Hz)
% N_b: Number of bands


if(nargin<5); plotcode = 0; end;

% Selection of which bands are used
switch bands
    case 1
        % Bark bands
        % Upper & lower limits & bandwidths
        omegaL = 2*pi/44100.*[20, 100, 200, 300, 400, 510, 630, 770, 920, 1080, 1270, 1480, 1720, 2000, 2320, 2700, 3150, 3700, 4400, 5300, 6400, 7700, 9500, 12000 15500]; % Lower cutoff frequency (rad)
        omegaU = 2*pi/44100.*[100, 200, 300, 400, 510, 630, 770, 920, 1080, 1270, 1480, 1720, 2000, 2320, 2700, 3150, 3700, 4400, 5300, 6400, 7700, 9500, 12000, 15500 20500]; % Upper cutoff frequency (rad)
        omegaB = omegaU-omegaL;                                 % Bandwidth (rad)
        omegaM = 2.*atan(sqrt(tan(omegaU/2).*tan(omegaL/2)));   % Optimized center frequencies (rad)
        f_M = round(fs.*omegaM'./(2*pi));                       % Optimized center frequency (Hz)
        N_b = length(omegaM);                                   % Number of bands
    case 2
        % 1/3 octave bands
        % Upper & lower limits & bandwidths
        omegaL = 2*pi/fs.*[28.2 35.5 44.7 56.2 70.8 89.1 112 141 178 224 282 355 447 562 708 891 1120 1410 1780 2240 2820 3550 4470 5620 7080 8910 11200 14100 17800];
        omegaU = 2*pi/fs.*[35.5 44.7 56.2 70.8 89.1 112 141 178 224 282 355 447 562 708 891 1120 1410 1780 2240 2820 3550 4470 5620 7080 8910 11200 14100 17800 22400];
        omegaB = omegaU-omegaL;                                 % Bandwidth (rad)
        omegaM = 2.*atan(sqrt(tan(omegaU/2).*tan(omegaL/2)));   % Optimized center frequencies (rad)
        if(fs<48000)
            omegaM = omegaM(1:end-1);
        end
        f_M = round(fs.*omegaM'./(2*pi));                       % Optimized center frequency (Hz)
        N_b = length(omegaM);                                   % Number of bands
    case 3
        % octave bands
        % Upper & lower limits % bandwidths
        omegaL = 2*pi/fs.*[28.2 44.7 89.1 178 355 708 1410 2820 5620 11200];
        omegaU = 2*pi/fs.*[44.7 89.1 178 355 708 1410 2820 5620 11200 22400];
        omegaB = omegaU-omegaL;                                 % Bandwidth (rad)
        omegaM = 2.*atan(sqrt(tan(omegaU/2).*tan(omegaL/2)));   % Optimized center frequencies (rad)
        if(fs<48000)
            omegaM = omegaM(1:end-1);
        end
        f_M = round(fs.*omegaM'./(2*pi));                       % Optimized center frequency (Hz)
        N_b = length(omegaM);                                   % Number of bands
end

% Input Parameters
input = [1 zeros(1,16383)]; % Impulse, length 2^14
g_dB = g;                   % Gains in dB
g = 10.^(g_dB./20);         % Desired gain
M = 2*N_m;                  % M/2 = Number of 4th-order filter sections, order N = 2*M = 4*N_m

% Initialize variables
len = length(input);        % Number of output samples to be computed
x = input;                  % Input signal
y = zeros(1,len);           % Initialize output signal array


% Loop for cascading the frequency bands

for b=1:N_b
    
    % Coefficients used in the 4th-order section
    K = (1/(g(b).^(1/(2*M(b))))) * tan(omegaB(b)/2);    % Calculation of K
    V = g(b)^(1/M(b)) - 1;                              % Calculation of V (Note that -1 is missing in some versions of the DAFx-06 paper and poster!)
                                       
    a = cos(omegaM(b));                                 % Allpass filter coefficient (the same for A1 and A2)


    % Loop for cascading 4th-order structures
    for m=1:(M(b)/2)
    
        alpham = (1/2 - (2*m-1)/(2*M(b)))*pi;           % Calculation of alpha_m
        cm = cos(alpham);                               % Calculation of c_m
        a0m = 1/(1+2*K*cm+K^2);                         % Calculation of a_{0,m}^-1
    
        w11 = 0; w12 = 0; w21 = 0; w22 = 0;             % Initialize state variables (unit delay elements)
        v1 = 0; v2 = 0;                                 % Initialize intermediate variables
        a1out = 0; a2out = 0;                           % Initialize output variables of the allpass filters

        % Compute output samples y(n) in a loop
        for n = 1:length(x),
            % Allpass filter A1
            a1out = a*(w11 + a*w12) - w12;  % Output sample of allpass filter A1
            %
            % Allpass filter A2
            a2out = a*(w21 + a*w22) - w22;  % Output sample of allpass filter A2
            %
            % Intermediate variables of the 4th-order section
            v1 = (a2out - 2*a1out) + K*(2*(-cm)*a2out + K*(2*a1out + a2out));
            v2 = a2out + 2*a1out;
            %
            % Output signal of the 4th-order section
            y(n) = x(n) + V*2*(-cm)*(-a0m*(K*x(n) - v1) + a2out) + ...
                V*(2+V)*K*(a0m*(K*x(n) - v1) + v2);
            %
            % Update state variables
            w12 = w11 + a*w12;                          % Move data inside allpass filter A1
            w11 = a0m*(K*x(n) - v1);                    % Move data through the unit delay into A1
            w22 = w21 + a*w22;                          % Move data inside allpass filter A2
            w21 = a1out;                                % Move data through the unit delay into A2
        end
        x = y;  % Set the previous output to be the next input
    end
    
end

out = y; 

if(plotcode)
    % Calculate spectrum
    len_sp = length(y);
    siz = floor(len/2);
    Y = fft(y);
    Ysp = 20*log10(abs(Y(1:siz)));
    dfreq = fs/len;
    % Plot spectrum
    figure(1)
    frq = linspace(0,dfreq*(siz-1),siz);
    p = semilogx(frq,Ysp,'k');
    hold on;
    for i = 1:N_b
        plot([fs*omegaU(i)/(2*pi) fs*omegaU(i)/(2*pi)], [-80 80],'--k');
    end
    hold off;
    % Plot style
    xlim([30 fs/2])
    % ylim calculation
    gmax = max(g_dB);
    gmin = min(g_dB);
    ylim([min(0,gmin-10) max(0,gmax+10)])
    set(gca,'XTick',[30 100 300 1000 3000 10000]);
    set(gca,'XTickLabel',{'30','100','300','1k','3k','10k'},'Fontname','Times','Fontsize',18)
    xlabel('Frequency (Hz)')
    ylabel('Magnitude (dB)')
end

